
macro_rules! __impl_from_into_for_rational {
    ($impl_name:ident, $raw_type:ty, $frac_bits:expr, $from_type:ty) => {
        impl From<$from_type> for $impl_name {
            fn from(value: $from_type) -> Self {
                let integral = (value as $raw_type).shl($frac_bits);
                let fraction = (value.fract() * 1.shl($frac_bits) as $from_type) as $raw_type;
                $impl_name(integral | fraction)
            }
        }

        impl Into<$from_type> for $impl_name {
            fn into(self) -> $from_type {

                let integral = self.0.shr($frac_bits);
                let fraction = (self.0.bitand(FRAC_MASK)) as $from_type / (1 as $raw_type).shl($frac_bits) as $from_type;

                integral as $from_type + fraction
            }
        }
    }
}

macro_rules! impl_fixed_point {
    ($impl_name:ident, $raw_type:ty, $frac_bits:expr) => {
        mod $impl_name {
            use std::ops;
            use std::ops::{Add, BitAnd, Div, Mul, Shl, Shr, Sub};

            #[derive(Copy, Clone, PartialOrd, PartialEq, Eq, Ord)]
            pub struct $impl_name($raw_type);

            pub const ZERO: $impl_name = $impl_name(0);

            const FRAC_MASK: $raw_type = (-1) << (std::mem::size_of::<$raw_type>() * 8 - $frac_bits);

            impl From<$raw_type> for $impl_name {
                fn from(value: $raw_type) -> Self {
                    $impl_name(value.shl($frac_bits))
                }
            }

            __impl_from_into_for_rational!($impl_name, $raw_type, $frac_bits, f32);
            __impl_from_into_for_rational!($impl_name, $raw_type, $frac_bits, f64);

            impl Into<$raw_type> for $impl_name {
                fn into(self) -> $raw_type {
                    self.0.shr($frac_bits)
                }
            }

            impl Add<$impl_name> for $impl_name {
                type Output = $impl_name;

                fn add(self, rhs: $impl_name) -> Self::Output {
                    $impl_name(self.0 + rhs.0)
                }
            }

            impl Sub<$impl_name> for $impl_name {
                type Output = $impl_name;

                fn sub(self, rhs: $impl_name) -> Self::Output {
                    $impl_name(self.0 - rhs.0)
                }
            }

            impl Mul<$impl_name> for $impl_name {
                type Output = $impl_name;

                fn mul(self, rhs: $impl_name) -> Self::Output {
                    $impl_name(self.0.shr($frac_bits / 2) * rhs.0.shr($frac_bits / 2))
                }
            }

            impl Div<$impl_name> for $impl_name {
                type Output = $impl_name;

                fn div(self, rhs: $impl_name) -> Self::Output {
                    $impl_name((self.0 / rhs.0.shr($frac_bits / 2 as $raw_type)).shl($frac_bits / 2))
                }
            }

            impl Default for $impl_name {
                fn default() -> Self {
                    ZERO
                }
            }

            impl $impl_name {

                pub fn integral(&self) -> $raw_type {
                    self.0.shr($frac_bits)
                }

                pub fn fraction_bits(&self) -> $raw_type {
                    self.0
                }
            }
        }
    }
}

#[cfg(test)]
mod tests {
    use crate::tests::LowpFixed::LowpFixed;
    use super::*;

    impl_fixed_point!(LowpFixed, i32, 16);
    impl_fixed_point!(HighpFixed, i64, 32);

    #[test]
    fn test() {
        let a = HighpFixed::ZERO;
        let b =  HighpFixed::HighpFixed::from(3);

        let c = a * a / b;

        println!("{}", Into::<f32>::into(c));
    }
}
